/*============================================================================*/
/*                                                                            */
/*                                                                            */
/*                              reLOC 0.09-vegas                              */
/*                                                                            */
/*                   (C) Copyright 2011-2013 Pavel Surynek                    */
/*            http://www.surynek.com | <pavel.surynek@mff.cuni.cz>            */
/*                                                                            */
/*                                                                            */
/*============================================================================*/
// compress.h / 0.09-vegas_092
/*----------------------------------------------------------------------------*/
//
// Compression tools for relocation problem solutions.
//
/*----------------------------------------------------------------------------*/


#ifndef __COMPRESS_H__
#define __COMPRESS_H__

#include <pthread.h>

#include <vector>

#include "types.h"
#include "result.h"


using namespace std;
using namespace sReloc;


/*----------------------------------------------------------------------------*/

namespace sReloc
{


/*----------------------------------------------------------------------------*/
// sMultirobotSolutionCompressor

    class sMultirobotSolutionCompressor
    {
    public:
	static const int MAKESPAN_UNDEFINED = -1;
	static const int MINISAT_TIMEOUT_UNDEFINED = -1;
	static const int TOTAL_TIMEOUT_UNDEFINED = -1;
	static const int THREAD_ID_UNDEFINED = -1;

	static const sString CNF_INVERSE_FILENAME_PREFIX;
	static const sString CNF_ADVANCED_FILENAME_PREFIX;
	static const sString CNF_DIFFERENTIAL_FILENAME_PREFIX;
	static const sString CNF_BIJECTION_FILENAME_PREFIX;
	static const sString CNF_HEURISTIC_ADVANCED_FILENAME_PREFIX;
	static const sString CNF_HEURISTIC_DIFFERENTIAL_FILENAME_PREFIX;
	static const sString CNF_HEURISTIC_BIJECTION_FILENAME_PREFIX;
	static const sString CNF_BITWISE_FILENAME_PREFIX;
	static const sString CNF_FLOW_FILENAME_PREFIX;
	static const sString CNF_MATCHING_FILENAME_PREFIX;
	static const sString CNF_HEURISTIC_MATCHING_FILENAME_PREFIX;

	static const sString OUTPUT_INVERSE_FILENAME_PREFIX;
	static const sString OUTPUT_ADVANCED_FILENAME_PREFIX;
	static const sString OUTPUT_DIFFERENTIAL_FILENAME_PREFIX;
	static const sString OUTPUT_BIJECTION_FILENAME_PREFIX;
	static const sString OUTPUT_HEURISTIC_ADVANCED_FILENAME_PREFIX;
	static const sString OUTPUT_HEURISTIC_DIFFERENTIAL_FILENAME_PREFIX;
	static const sString OUTPUT_HEURISTIC_BIJECTION_FILENAME_PREFIX;
	static const sString OUTPUT_BITWISE_FILENAME_PREFIX;
	static const sString OUTPUT_FLOW_FILENAME_PREFIX;
	static const sString OUTPUT_MATCHING_FILENAME_PREFIX;
	static const sString OUTPUT_HEURISTIC_MATCHING_FILENAME_PREFIX;

    public:
	enum Encoding
	{
	    ENCODING_UNDEFINED,
	    ENCODING_UNUSED,
	    ENCODING_INVERSE,
	    ENCODING_ADVANCED,
	    ENCODING_DIFFERENTIAL,
	    ENCODING_BIJECTION,
	    ENCODING_HEURISTIC_ADVANCED,
	    ENCODING_HEURISTIC_DIFFERENTIAL,
	    ENCODING_HEURISTIC_BIJECTION,
	    ENCODING_PUZZLE,
	    ENCODING_BITWISE,
	    ENCODING_FLOW,
	    ENCODING_MATCHING,
	    ENCODING_HEURISTIC_MATCHING
	};

	typedef std::vector<sRobotArrangement> Arrangements_vector;

	struct CompressionRecord
	{
	    CompressionRecord();
	    CompressionRecord(int start_step, int final_step);

	    CompressionRecord(int                        makespan,
			      int                        start_step,
			      int                        final_step,
			      CompressionRecord         *left_component,
			      CompressionRecord         *right_component,
			      const sMultirobotSolution &compressed_sub_solution);

	    int m_makespan;
	    int m_start_step;
	    int m_final_step;

	    CompressionRecord *m_left_component;
	    CompressionRecord *m_right_component;

	    sMultirobotSolution m_compressed_sub_solution;
	};

	typedef std::vector<std::vector<CompressionRecord> > CompressionMatrix_2d_vector;

	struct ProcessingArgument_IN
	{
	    ProcessingArgument_IN();
	    ProcessingArgument_IN(const sRobotArrangement   &initial_arrangement,
				  const sMultirobotSolution &original_solution,
				  const sUndirectedGraph    &environment);

	    int m_start_step;
	    int m_final_step;

	    sRobotArrangement m_initial_arrangement;
	    sMultirobotSolution m_original_solution;
	    const sUndirectedGraph *m_environment;
	};

	struct ProcessingArgument_OUT
	{
	    ProcessingArgument_OUT();

	    sResult m_result;
	    sMultirobotSolution m_processed_solution;
	};

	struct ProcessingArgument_CTX
	{
	    ProcessingArgument_CTX();
	    ProcessingArgument_CTX(sMultirobotSolutionCompressor &compressor);

	    sMultirobotSolutionCompressor *m_compressor;
	    int m_thread_id;
	    pthread_t m_pthread_handle;
	};

	struct ProcessingArgument
	{
	    ProcessingArgument();
	    ProcessingArgument(const ProcessingArgument_CTX &ctx_arg, const ProcessingArgument_IN &in_arg);

	    ProcessingArgument_CTX m_ctx_arg;
	    ProcessingArgument_IN m_in_arg;
	    ProcessingArgument_OUT m_out_arg;
	};

	typedef std::vector<ProcessingArgument> ProcessingArguments_vector;

	struct ProcessedSolutionRecord
	{
	    ProcessedSolutionRecord();
	    ProcessedSolutionRecord(int start_step, int final_step, const sMultirobotSolution &short_solution);

	    bool operator==(const ProcessedSolutionRecord &processed_record) const;
	    bool operator<(const ProcessedSolutionRecord &processed_record) const;
	    
	    int m_start_step;
	    int m_final_step;

	    sMultirobotSolution m_processed_solution;
	};

	typedef std::set<ProcessedSolutionRecord, std::less<ProcessedSolutionRecord> > ProcessedSolutionRecords_set;

	struct AttemptDatabaseRecord
	{
	    AttemptDatabaseRecord();
	    AttemptDatabaseRecord(const sRobotArrangement &initial_arrangement, const sRobotArrangement &goal_arrangement);

	    bool operator==(const AttemptDatabaseRecord &attempt_record) const;
	    bool operator<(const AttemptDatabaseRecord &attempt_record) const;

	    sRobotArrangement m_initial_arrangement;
	    sRobotArrangement m_goal_arrangement;
	};

	typedef std::set<AttemptDatabaseRecord, std::less<AttemptDatabaseRecord> > AttemptDatabaseRecords_set;

    public:
	sMultirobotSolutionCompressor(const sString &minisat_path,
				      int            minisat_timeout       = MINISAT_TIMEOUT_UNDEFINED,
				      int            total_timeout         = TOTAL_TIMEOUT_UNDEFINED,
				      int            makespan_upper_bound  = sDEFAULT_MINISAT_UPPER_BOUND,
	                              int            N_parallel_Threads    = sDEFAULT_N_PARALLEL_THREADS,
				      Encoding       encoding              = ENCODING_INVERSE);

	sMultirobotSolutionCompressor(const sMultirobotSolutionCompressor &solution_compressor);
	const sMultirobotSolutionCompressor& operator=(const sMultirobotSolutionCompressor &solution_compressor);

	int calc_MakespanLowerBound(const sRobotArrangement                     &start_arrangement,
				    const sRobotArrangement                     &final_arrangement,
				    const sUndirectedGraph::Distances_2d_vector &all_pairs_Distances);

	int calc_MakespanLowerBound(const sRobotArrangement                     &start_arrangement,
				    const sRobotGoal                            &robot_goal,
				    const sUndirectedGraph::Distances_2d_vector &all_pairs_Distances);

	sResult compute_OptimalMakespan(const sRobotArrangement &start_arrangement,
					const sRobotGoal        &final_arrangement,
					const sUndirectedGraph  &environment,
					int                      makespan_upper_bound,
					int                     &optimal_makespan,
					int                      thread_id = THREAD_ID_UNDEFINED);

	sResult compute_OptimalMakespan_(const sRobotArrangement &start_arrangement,
					 const sRobotGoal        &final_arrangement,
					 const sUndirectedGraph  &environment,
					 int                      makespan_upper_bound,
					 int                     &optimal_makespan,
					 int                      thread_id = THREAD_ID_UNDEFINED);

	sResult compute_OptimalMakespan(const sRobotArrangement &start_arrangement,
					const sRobotGoal        &final_arrangement,
					const sUndirectedGraph  &environment,
					int                      makespan_lower_bound,
					int                      makespan_upper_bound,
					int                     &optimal_makespan,
					int                      thread_id = THREAD_ID_UNDEFINED);

	sResult compute_OptimalMakespan(const sRobotArrangement           &start_arrangement,
					const sRobotGoal                  &final_arrangement,
					const sUndirectedGraph            &environment,
					int                                makespan_upper_bound,
					int                               &optimal_makespan,
					sMultirobotEncodingContext_CNFsat &final_encoding_context,
					int                                thread_id = THREAD_ID_UNDEFINED);

	sResult compute_OptimalMakespan_(const sRobotArrangement           &start_arrangement,
					 const sRobotGoal                  &final_arrangement,
					 const sUndirectedGraph            &environment,
					 int                                makespan_upper_bound,
					 int                               &optimal_makespan,
					 sMultirobotEncodingContext_CNFsat &final_encoding_context,
					 int                                thread_id = THREAD_ID_UNDEFINED);

	sResult compute_OptimalMakespan(const sRobotArrangement           &start_arrangement,
					const sRobotGoal                  &final_arrangement,
					const sUndirectedGraph            &environment,
					int                                makespan_lower_bound,
					int                                makespan_upper_bound,
					int                               &optimal_makespan,
					sMultirobotEncodingContext_CNFsat &final_encoding_context,
					int                                thread_id = THREAD_ID_UNDEFINED);

	sResult compute_OptimalSolution(const sRobotArrangement &start_arrangement,
					const sRobotGoal        &final_arrangement,
					const sUndirectedGraph  &environment,
					int                      makespan_upper_bound,
					int                     &optimal_makespan,
					sMultirobotSolution     &optimal_solution,
					int                      thread_id = THREAD_ID_UNDEFINED);

	sResult compute_OptimalSolution_(const sRobotArrangement &start_arrangement,
					 const sRobotGoal        &final_arrangement,
					 const sUndirectedGraph  &environment,
					 int                      makespan_upper_bound,
					 int                     &optimal_makespan,
					 sMultirobotSolution     &optimal_solution,
					 int                      thread_id = THREAD_ID_UNDEFINED);

	sResult compute_OptimalSolution(const sRobotArrangement &start_arrangement,
					const sRobotGoal        &final_arrangement,
					const sUndirectedGraph  &environment,
					int                      makespan_lower_bound,
					int                      makespan_upper_bound,
					int                     &optimal_makespan,
					sMultirobotSolution     &optimal_solution,
					int                      thread_id = THREAD_ID_UNDEFINED);


/*----------------------------------------------------------------------------*/

	sResult compute_SuboptimalMakespan(const sRobotArrangement &start_arrangement,
					   const sRobotArrangement &final_arrangement,
					   const sUndirectedGraph  &environment,
					   int                      makespan_lower_bound,
					   int                      makespan_upper_bound,
					   int                     &suboptimal_makespan,
					   int                      thread_id = THREAD_ID_UNDEFINED);

	sResult compute_SuboptimalMakespan(const sRobotArrangement           &start_arrangement,
					   const sRobotArrangement           &final_arrangement,
					   const sUndirectedGraph            &environment,
					   int                                makespan_lower_bound,
					   int                                makespan_upper_bound,
					   int                               &suboptimal_makespan,
					   sMultirobotEncodingContext_CNFsat &final_encoding_context,
					   int                                thread_id = THREAD_ID_UNDEFINED);
	
	sResult compute_SuboptimalSolution(const sRobotArrangement &start_arrangement,
					   const sRobotArrangement &final_arrangement,
					   const sUndirectedGraph  &environment,
					   int                      makespan_lower_bound,
					   int                      makespan_upper_bound,
					   int                     &suboptimal_makespan,
					   sMultirobotSolution     &suboptimal_solution,
					   int                      thread_id = THREAD_ID_UNDEFINED);


/*----------------------------------------------------------------------------*/

	sResult extract_ComputedInverseSolution(const sRobotArrangement                 &start_arrangement,
						const sUndirectedGraph                  &environment,
						int                                      computed_makespan,
						const sMultirobotEncodingContext_CNFsat &final_encoding_context,
						sMultirobotSolution                     &computed_solution,
						int                                      thread_id = THREAD_ID_UNDEFINED);

	sResult extract_ComputedAdvancedSolution(const sRobotArrangement                 &start_arrangement,
						 const sUndirectedGraph                  &environment,
						 int                                      computed_makespan,
						 const sMultirobotEncodingContext_CNFsat &final_encoding_context,
						 sMultirobotSolution                     &computed_solution,
						 int                                      thread_id = THREAD_ID_UNDEFINED);

	sResult extract_ComputedDifferentialSolution(const sRobotArrangement                 &start_arrangement,
						     const sUndirectedGraph                  &environment,
						     int                                      computed_makespan,
						     const sMultirobotEncodingContext_CNFsat &final_encoding_context,
						     sMultirobotSolution                     &computed_solution,
						     int                                      thread_id = THREAD_ID_UNDEFINED);

	sResult extract_ComputedBijectionSolution(const sRobotArrangement                 &start_arrangement,
						  const sUndirectedGraph                  &environment,
						  int                                      computed_makespan,
						  const sMultirobotEncodingContext_CNFsat &final_encoding_context,
						  sMultirobotSolution                     &computed_solution,
						  int                                      thread_id = THREAD_ID_UNDEFINED);

	sResult extract_ComputedHeuristicAdvancedSolution(const sRobotArrangement                 &start_arrangement,
							  const sUndirectedGraph                  &environment,
							  int                                      computed_makespan,
							  const sMultirobotEncodingContext_CNFsat &final_encoding_context,
							  sMultirobotSolution                     &computed_solution,
							  int                                      thread_id = THREAD_ID_UNDEFINED);

	sResult extract_ComputedHeuristicDifferentialSolution(const sRobotArrangement                 &start_arrangement,
							      const sUndirectedGraph                  &environment,
							      int                                      computed_makespan,
							      const sMultirobotEncodingContext_CNFsat &final_encoding_context,
							      sMultirobotSolution                     &computed_solution,
							      int                                      thread_id = THREAD_ID_UNDEFINED);
	
	sResult extract_ComputedHeuristicBijectionSolution(const sRobotArrangement                 &start_arrangement,
							   const sUndirectedGraph                  &environment,
							   int                                      computed_makespan,
							   const sMultirobotEncodingContext_CNFsat &final_encoding_context,
							   sMultirobotSolution                     &computed_solution,
							   int                                      thread_id = THREAD_ID_UNDEFINED);

	sResult extract_ComputedBitwiseSolution(const sRobotArrangement                 &start_arrangement,
						const sUndirectedGraph                  &environment,
						int                                      computed_makespan,
						const sMultirobotEncodingContext_CNFsat &final_encoding_context,
						sMultirobotSolution                     &computed_solution,
						int                                      thread_id = THREAD_ID_UNDEFINED);

	sResult extract_ComputedFlowSolution(const sRobotArrangement                 &start_arrangement,
					     const sUndirectedGraph                  &environment,
					     int                                      computed_makespan,
					     const sMultirobotEncodingContext_CNFsat &final_encoding_context,
					     sMultirobotSolution                     &computed_solution,
					     int                                      thread_id = THREAD_ID_UNDEFINED);

	sResult extract_ComputedMatchingSolution(const sRobotArrangement                 &start_arrangement,
						 const sUndirectedGraph                  &environment,
						 int                                      computed_makespan,
						 const sMultirobotEncodingContext_CNFsat &final_encoding_context,
						 sMultirobotSolution                     &computed_solution,
						 int                                      thread_id = THREAD_ID_UNDEFINED);

	sResult extract_ComputedHeuristicMatchingSolution(const sRobotArrangement                 &start_arrangement,
							  const sUndirectedGraph                  &environment,
							  int                                      computed_makespan,
							  const sMultirobotEncodingContext_CNFsat &final_encoding_context,
							  sMultirobotSolution                     &computed_solution,
							  int                                      thread_id = THREAD_ID_UNDEFINED);


	sResult shorten_Solution(const sRobotArrangement   &initial_arrangement,
				 const sMultirobotSolution &original_solution,
				 const sUndirectedGraph    &environment,
				 sMultirobotSolution       &shortened_solution,
				 int                        thread_id = THREAD_ID_UNDEFINED);

	sResult shorten_Solution_mt(const sRobotArrangement   &initial_arrangement,
				    const sMultirobotSolution &original_solution,
				    const sUndirectedGraph    &environment,
				    sMultirobotSolution       &shortened_solution);

	sResult primeShorten_Solution(const sRobotArrangement   &initial_arrangement,
				       const sMultirobotSolution &original_solution,
				       const sUndirectedGraph    &environment,
				       sMultirobotSolution       &shortened_solution,
				       int                        thread_id = THREAD_ID_UNDEFINED);

	sResult primeShorten_Solution_mt(const sRobotArrangement   &initial_arrangement,
					  const sMultirobotSolution &original_solution,
					  const sUndirectedGraph    &environment,
					  sMultirobotSolution       &shortened_solution);

	sResult primeShorten_Solution_mt2(const sRobotArrangement   &initial_arrangement,
					   const sMultirobotSolution &original_solution,
					   const sUndirectedGraph    &environment,
					   sMultirobotSolution       &shortened_solution);

	sResult optimize_Solution(const sRobotArrangement   &initial_arrangement,
				  const sMultirobotSolution &original_solution,
				  const sUndirectedGraph    &environment,
				  sMultirobotSolution       &optimized_solution);

	sResult optimize_Solution_mt(const sRobotArrangement   &initial_arrangement,
				     const sMultirobotSolution &original_solution,
				     const sUndirectedGraph    &environment,
				     sMultirobotSolution       &optimized_solution);

	sResult primeOptimize_Solution(const sRobotArrangement   &initial_arrangement,
					const sMultirobotSolution &original_solution,
					const sUndirectedGraph    &environment,
					sMultirobotSolution       &optimized_solution);

	sResult primeOptimize_Solution_mt(const sRobotArrangement   &initial_arrangement,
					   const sMultirobotSolution &original_solution,
					   const sUndirectedGraph    &environment,
					   sMultirobotSolution       &optimized_solution);

	sResult compress_Solution(const sRobotArrangement   &initial_arrangement,
				  const sMultirobotSolution &original_solution,
				  const sUndirectedGraph    &environment,
				  sMultirobotSolution       &compressed_solution);

	sResult deflate_Solution(const sRobotArrangement   &initial_arrangement,
				 const sMultirobotSolution &original_solution,
				 const sUndirectedGraph    &environment,
				 sMultirobotSolution       &deflated_solution);

	sResult deflate_Solution_mt(const sRobotArrangement   &initial_arrangement,
				    const sMultirobotSolution &original_solution,
				    const sUndirectedGraph    &environment,
				    sMultirobotSolution       &deflated_solution);

	static void unfold_Solution(const sRobotArrangement &initial_arrangement, const sMultirobotSolution &solution, Arrangements_vector &unfolded_Arrangements);
	static bool verify_Unfolding(const sRobotArrangement &initial_arrangement, const sMultirobotSolution &solution, const sUndirectedGraph &graph);

	void to_Screen_arrangements(const Arrangements_vector &unfolded_Arrangements, const sString &indent = "") const;
	void to_Stream_arrangements(FILE *fw, const Arrangements_vector &unfolded_Arrangements, const sString &indent = "") const;

    private:
	int m_minisat_timeout;
	int m_total_timeout;
	int m_makespan_upper_bound;
	sString m_minisat_path;

	int m_N_parallel_Threads;
	Encoding m_encoding;

	AttemptDatabaseRecords_set m_attempt_Database;
    };


/*----------------------------------------------------------------------------*/
// Global functions

    void* s_shorten_Solution_mt(void *arg);
    void* s_prime_shorten_Solution_mt(void *arg);


/*----------------------------------------------------------------------------*/

} // namespace sReloc

#endif /* __COMPRESS_H__ */
